@@ -9,6 +9,7 @@ from sklearn.metrics import auc
 from sklearn import metrics
 from scipy.stats import t
 from scipy import stats
+from scipy.stats import ks_2samp
 
 
 def _binomial(p, d, n):
@@ -108,7 +109,13 @@ def binomial_test(data, ratings, default_flag, predicted_pd, alpha_level=0.05):
 
     # Transform input data into the required format
     df = data.groupby(ratings).agg({predicted_pd: "mean", default_flag: ["count", "sum", "mean"]}).reset_index()
-    df.columns = ["Rating class", "Predicted PD", "Total count", "Defaults", "Actual Default Rate"]
+    df.columns = [
+        "Rating class",
+        "Predicted PD",
+        "Total count",
+        "Defaults",
+        "Actual Default Rate",
+    ]
 
     # Calculate Binomial test outcome for each rating
     df["p_value"] = _binomial(df["Predicted PD"], df["Defaults"], df["Total count"])
@@ -785,7 +792,13 @@ def jeffreys_test(data, ratings, default_flag, predicted_pd, alpha_level=0.05):
 
     # Transform input data into the required format
     df = data.groupby(ratings).agg({predicted_pd: "mean", default_flag: ["count", "sum", "mean"]}).reset_index()
-    df.columns = ["Rating class", "Predicted PD", "Total count", "Defaults", "Actual Default Rate"]
+    df.columns = [
+        "Rating class",
+        "Predicted PD",
+        "Total count",
+        "Defaults",
+        "Actual Default Rate",
+    ]
 
     # Calculate Binomial test outcome for each rating
     df["p_value"] = _jeffreys(df["Predicted PD"], df["Defaults"], df["Total count"])
@@ -794,7 +807,7 @@ def jeffreys_test(data, ratings, default_flag, predicted_pd, alpha_level=0.05):
     return df
 
 
-def roc_auc(df, target, prediction):
+def roc_auc(data, target, prediction):
     """Compute Area ROC AUC from prediction scores.
 
     Note: this implementation can be used with binary, multiclass and
@@ -819,14 +832,43 @@ def roc_auc(df, target, prediction):
     https://github.com/scikit-learn/scikit-learn/blob/36958fb24/sklearn/metrics/_ranking.py#L47
     """
 
+    # Perform plausibility checks
+    assert all(x >= 0 and x <= 1 for x in data[target]), "Predicted PDs must be between 0% and 100%"
+    assert all(x >= 0 and x <= 1 for x in data[prediction]), "Predicted PDs must be between 0% and 100%"
+
+    return roc_auc_score(data[target], data[prediction])
+
+
+def gini(df, target, prediction):
+    """Compute Area ROC AUC from prediction scores.
+
+    todo
+    """
+
     # Perform plausibility checks
     assert all(x >= 0 and x <= 1 for x in df[target]), "Predicted PDs must be between 0% and 100%"
     assert all(x >= 0 and x <= 1 for x in df[prediction]), "Predicted PDs must be between 0% and 100%"
 
-    return roc_auc_score(df[target], df[prediction])
+    roc = roc_auc(df, target, prediction)
+
+    return roc * 2 - 1
 
 
-def clar(df, predicted_ratings, realised_outcomes):
+def kolmogorov_smirnov_stat(df, target, prediction):
+    """Compute Area ROC AUC from prediction scores.
+    todo
+    """
+
+    # Perform plausibility checks
+    assert all(x >= 0 and x <= 1 for x in df[target]), "Predicted PDs must be between 0% and 100%"
+    assert all(x >= 0 and x <= 1 for x in df[prediction]), "Predicted PDs must be between 0% and 100%"
+
+    result = ks_2samp(df[target], df[prediction])
+
+    return result
+
+
+def cumulative_lgd_accuracy_ratio(df, predicted_ratings, realised_outcomes):
     """
     CLAR serves as a measure of ranking ability against LGD risk
     The cumulative LGD accuracy ratio (CLAR) curve can be treated as
@@ -862,7 +904,7 @@ def clar(df, predicted_ratings, realised_outcomes):
 
     for i, j in enumerate(list(set(df[predicted_ratings]))[::-1]):
         x = (df[predicted_ratings] == j).sum()
-        x_bucket = df.sort_values(by=realised_outcomes, ascending=False)[x_s[i]: x_s[i] + x]
+        x_bucket = df.sort_values(by=realised_outcomes, ascending=False)[x_s[i] : x_s[i] + x]
         x_value = x / len(df)
         y_value = (x_bucket[realised_outcomes] == j).sum() / len((x_bucket[realised_outcomes] == j))
         x_values.append(x_value)
@@ -873,9 +915,9 @@ def clar(df, predicted_ratings, realised_outcomes):
     new_yvalues = list(np.cumsum(y_values))
 
     model_auc = auc(new_xvalues, new_yvalues)
-    clar = 2 * model_auc
+    clar_value = 2 * model_auc
 
-    return clar
+    return clar_value
 
 
 def loss_capture_ratio(ead, predicted_ratings, realised_outcomes):
@@ -917,7 +959,11 @@ def loss_capture_ratio(ead, predicted_ratings, realised_outcomes):
     """
 
     # Create a dataframe
-    frame = {"ead": ead, "predicted_ratings": predicted_ratings, "realised_outcomes": realised_outcomes}
+    frame = {
+        "ead": ead,
+        "predicted_ratings": predicted_ratings,
+        "realised_outcomes": realised_outcomes,
+    }
     df = pd.DataFrame(frame)
 
     # Prepare data
@@ -937,12 +983,12 @@ def loss_capture_ratio(ead, predicted_ratings, realised_outcomes):
     auc_curve2 = auc([i for i in range(len(df3))], df3.cumulative_loss_capture_percentage)
     random_auc2 = 0.5 * len(df3) * 1
 
-    loss_capture_ratio = (auc_curve1 - random_auc1) / (auc_curve2 - random_auc2)
+    lcr = (auc_curve1 - random_auc1) / (auc_curve2 - random_auc2)
 
-    return loss_capture_ratio
+    return lcr
 
 
-def bayesian_error_rate(default_flag, prob_default):
+def bayesian_error_rate(df, default_flag, prob_default):
     """
     BER is the proportion of the whole sample that is misclassified
     when the rating system is in optimal use. For a perfect rating model,
@@ -952,7 +998,7 @@ def bayesian_error_rate(default_flag, prob_default):
     The Bayesian error rate specifies the minimum probability of error if
     the rating system or score function under consideration is used for a
     yes/no decision whether a borrower will default or not. The error can
-    be estimated parametrically, e.g. assuming normal score distributions,
+    be estimated parametrically, e.g. assuming noFrmal score distributions,
     or non-parametrically, for instance with kernel density estimation methods.
     If parametric estimation is applied, the distributional assumptions have
     to be carefully checked. Non-parametric estimation will be critical if
@@ -984,11 +1030,11 @@ def bayesian_error_rate(default_flag, prob_default):
     -0.47140452079103173
     """
 
-    frame = {"default_flag": default_flag, "prob_default": prob_default}
+    # frame = {"default_flag": default_flag, "prob_default": prob_default}
 
-    df = pd.DataFrame(frame)
+    # df = pd.DataFrame(frame)
 
-    fpr, tpr, thresholds = metrics.roc_curve(df["default_flag"], df["prob_default"])
+    fpr, tpr, thresholds = metrics.roc_curve(df[default_flag], df[prob_default])
     roc_curve_df = pd.DataFrame({"c": thresholds, "hit_rate": tpr, "false_alarm_rate": fpr})
 
     p_d = df.default_flag.sum() / len(df)
@@ -998,7 +1044,7 @@ def bayesian_error_rate(default_flag, prob_default):
     return round(min(roc_curve_df["ber"]), 3)
 
 
-def calc_iv(df, feature, target, pr=0):
+def information_value(df, feature, target, pr=0):
     """
     A numerical value that quantifies the predictive power of an independent
     variable in capturing the binary dependent variable.
@@ -1065,7 +1111,7 @@ def calc_iv(df, feature, target, pr=0):
     return data, iv
 
 
-def lgd_t_test(df, observed_LGD_col, expected_LGD_col, verbose=False):
+def lgd_t_test(df, observed_lgd, expected_lgd, level="portfolio", segment_col=None):
     """t-test for the Null hypothesis that estimated LGD is greater than true LGD
     Parameters
     ----------
@@ -1075,6 +1121,8 @@ def lgd_t_test(df, observed_LGD_col, expected_LGD_col, verbose=False):
         name of column with observed LGD values
     expected_LGD_col: string
         name of column with expected LGD values
+    level: string
+        'portfolio' (single comparison) or 'segment' level (multiple comparisons)
     verbose: boolean
         if true, results and interpretation are printed
     Returns
@@ -1104,51 +1152,88 @@ def lgd_t_test(df, observed_LGD_col, expected_LGD_col, verbose=False):
     # Checking for any missing data
     if df.empty:
         raise TypeError("No data provided!")
-    if observed_LGD_col is None:
+    if observed_lgd is None:
         raise TypeError("No column name for observed LGDs provided")
-    if expected_LGD_col is None:
+    if expected_lgd is None:
         raise TypeError("No column name for expected LGDs provided")
 
-    # Checking that the correct datatype
-    if not isinstance(observed_LGD_col, str):
-        raise TypeError("observed_LGD_col not of type string")
-    if not isinstance(expected_LGD_col, str):
-        raise TypeError("expected_LGD_col not of type string")
-
-    # Check if the correct column names have been provided
-    if observed_LGD_col not in df.columns:
-        raise ValueError("{} not a column in the df".format(observed_LGD_col))
-    if expected_LGD_col not in df.columns:
-        raise ValueError("{} not a column in the df".format(expected_LGD_col))
-
     # Check the data for missing values
-    if df[observed_LGD_col].hasnans:
-        raise ValueError("Missing values in {}".format(observed_LGD_col))
-    if df[expected_LGD_col].hasnans:
-        raise ValueError("Missing values in {}".format(expected_LGD_col))
+    if df[observed_lgd].hasnans:
+        raise ValueError("Missing values in {}".format(observed_lgd))
+    if df[expected_lgd].hasnans:
+        raise ValueError("Missing values in {}".format(expected_lgd))
 
-    N = len(df)
-    LGD = df[observed_LGD_col]
-    pred_LGD = df[expected_LGD_col]
-    error = LGD - pred_LGD
-    mean_error = error.mean()
-    num = np.sqrt(N) * mean_error
-    lgd_s2 = ((error - mean_error) ** 2).sum() / (N - 1)
-    t_stat = num / np.sqrt(lgd_s2)
-    p_value = 1 - t.cdf(t_stat, df=N - 1)
+    results = []
+    if level == "pool":
+        
 
-    if verbose is True:
-        # print the results
-        print(
-            "t_stat=%.3f, LGD.mean=%.3f,pred_LGD.mean=%.3f,N=%d, s2=%.3f, p=%.3f"
-            % (t_stat, pred_LGD.mean(), LGD.mean(), N, lgd_s2, p_value)
+        for segment in df[segment_col].unique():
+            df_segment = df[df[segment_col] == segment]
+
+            length = len(df_segment)
+            obs_lgd = df_segment[observed_lgd]
+            pred_lgd = df_segment[expected_lgd]
+            error = obs_lgd - pred_lgd
+            mean_error = error.mean()
+            num = np.sqrt(length) * mean_error
+            lgd_s2 = ((error - mean_error) ** 2).sum() / (length - 1)
+            t_stat = num / np.sqrt(lgd_s2)
+            p_value = 1 - t.cdf(t_stat, df=length - 1)
+
+            results.append(
+                [
+                    segment,
+                    length,
+                    obs_lgd.mean(),
+                    pred_lgd.mean(),
+                    lgd_s2,
+                    mean_error,
+                    t_stat,
+                    p_value,
+                ]
+            )
+
+    else:
+        segment = "None"
+        length = len(df)
+        obs_lgd = df[observed_lgd]
+        pred_lgd = df[expected_lgd]
+        error = obs_lgd - pred_lgd
+        mean_error = error.mean()
+        num = np.sqrt(length) * mean_error
+        lgd_s2 = ((error - mean_error) ** 2).sum() / (length - 1)
+        t_stat = num / np.sqrt(lgd_s2)
+        p_value = 1 - t.cdf(t_stat, df=length - 1)
+
+        results.append(
+            [
+                segment,
+                length,
+                obs_lgd.mean(),
+                pred_lgd.mean(),
+                lgd_s2,
+                mean_error,
+                t_stat,
+                p_value,
+            ]
         )
-        if p_value <= 0.05:
-            print("P-value <= 5%, therefore, H0 is rejected.")
-        elif p_value > 0.05:
-            print("P-value > 5%, therefore, H0 fails to be rejected.")
 
-    return N, LGD.mean(), pred_LGD.mean(), t_stat, lgd_s2, p_value
+        # from list of lists to dataframe
+    results_df = pd.DataFrame(
+        results,
+        columns=[
+            "segment",
+            "N",
+            "realised_lgd_mean",
+            "pred_lgd_mean",
+            "s2",
+            "mean_error",
+            "t_stat",
+            "p_value",
+        ],
+    )
+
+    return results_df  # todo: results do not make sense
 
 
 def migration_matrix_stability(df, initial_ratings_col, final_ratings_col):
@@ -1214,11 +1299,11 @@ def migration_matrix_stability(df, initial_ratings_col, final_ratings_col):
                 z_ij = np.nan
 
             z_df.iloc[i - 1, j - 1] = z_ij
-    phi_df = z_df.apply(lambda x: x.apply(lambda y: norm.cdf(y)))
+    phi_df = z_df.apply(lambda x: x.apply(norm.cdf))
     return z_df, phi_df
 
 
-def psi(data, bin_flag, variable):  # todo: expected vs actual
+def population_stability_index(data, bin_flag, variable):  # todo: expected vs actual
     """Calculate the PSI for a single variable
     Args:
         expected_array: numpy array of original values
@@ -1364,7 +1449,7 @@ def somersd(array_1, array_2, alternative="two-sided"):
     return stats.somersd(array_1, array_2, alternative="two-sided")
 
 
-def spearman_corr(array_1, array_2, alternative="two-sided"):
+def spearman_correlation(array_1, array_2):
     """
     Calculate a Spearman correlation coefficient with associated p-value.
     This is a wrapper around scipy.stats.spearmanr function.
@@ -1425,7 +1510,7 @@ def spearman_corr(array_1, array_2, alternative="two-sided"):
     return stats.spearmanr(array_1, array_2, alternative="two-sided")
 
 
-def pearson_corr(array_1, array_2):
+def pearson_correlation(array_1, array_2):
     """
     Calculate a Pearson correlation coefficient with associated p-value.
     This is a wrapper around scipy.stats.pearsonr function.
@@ -1461,4 +1546,175 @@ def pearson_corr(array_1, array_2):
         Probability and Statistics Tables and Formulae. Chapman & Hall: New
     """
 
-    pass
+    return stats.spearmanr(array_1, array_2)
+
+
+def migration_matrices_statistics(df, period_1_ratings, period_2_ratings):
+    """
+    The objective of this validation tool is to analyse the migration of customers across
+    rating grades during the relevant observation period
+    """
+    a = df[period_1_ratings]
+    b = df[period_2_ratings]
+
+    n_ij = pd.crosstab(a, b)
+    p_ij = pd.crosstab(a, b, normalize="index")
+
+    mnormu = 0
+    K = len(set(df["period_1_ratings"]))
+    for i in range(1, K - 1 + 1):
+        for j in range(i + 1, K + 1):
+            cac = p_ij.iloc[i - 1 : i, i:].sum(axis=1).values[0]
+            b = n_ij.sum(axis=1).values[i - 1]
+            a = max(i - K, i - 1)
+            mnormu += a * b * cac
+
+    mnorml = 0
+    K = len(set(df["period_1_ratings"]))
+    for i in range(2, K + 1):
+        for j in range(1, i - 1 + 1):
+            coc = p_ij.iloc[i - 1 : i, i:].sum(axis=1).values[0]
+            b = n_ij.sum(axis=1).values[i - 1]
+            a = max(i - K, i - 1)
+            mnorml += a * b * coc
+
+    upper_mwb = 0
+    for i in range(1, K - 1 + 1):
+        for j in range(i + 1, K + 1):
+            upper_mwb += abs(i - j) * n_ij.sum(axis=1).values[i - 1] * p_ij.iloc[i - 1, j - 1]
+    upper_mwb = (1 / mnormu) * upper_mwb
+
+    lower_mwb = 0
+    for i in range(2, K + 1):
+        for j in range(1, i - 1 + 1):
+            lower_mwb += abs(i - j) * n_ij.sum(axis=1).values[i - 1] * p_ij.iloc[i - 1, j - 1]
+    lower_mwb = (1 / mnorml) * lower_mwb
+
+    return upper_mwb, lower_mwb
+
+
+def _entropy(data, realised_pd, count):
+    """
+    CIER measures the ratio of distance between Unconditional and Conditional Entropy to
+    Unconditional Entropy.
+    """
+
+    # prepare data
+    data["perc"] = data[count] / data[count].sum()
+
+    # unconditional entropy
+    pd = (data[count] * data[realised_pd]).sum() / data[count].sum()
+
+    # Unconditional entropy
+    h0 = -(pd * np.log(pd) + (1 - pd) * np.log(1 - pd))
+
+    # Conditional entropy
+    data["hc"] = -(
+        data[realised_pd] * np.log(data[realised_pd]) + (1 - data[realised_pd]) * np.log(1 - data[realised_pd])
+    )
+    h1 = sum(data["perc"] * data["hc"])
+
+    return h0, h1
+
+
+def conditional_information_entropy_ratio(data, realised_pd, count):
+    """CIER measures the ratio of distance between Unconditional and
+    Conditional Entropy to Unconditional Entropy."""
+
+    h0, h1 = _entropy(data, realised_pd, count)
+
+    return (h0 - h1) / h0
+
+
+def kullback_leibler_dist(data, realised_pd, count):
+    """CIER measures the ratio of distance between Unconditional and
+    Conditional Entropy to Unconditional Entropy."""
+
+    h0, h1 = _entropy(data, realised_pd, count)
+
+    return h0 - h1
+
+
+def loss_shortfall(data, ead, predicted_lgd, realised_lgd):
+    """
+    Loss Shortfall is a measure of the difference between the expected loss and the
+    realised loss. It is a measure of the loss that would have been incurred if the
+    actual losses were equal to the expected losses.
+    """
+
+    estimated_losses = data[ead] * data[predicted_lgd]
+    realised_losses = data[ead] * data[realised_lgd]
+
+    return 1 - estimated_losses.sum() / realised_losses.sum()
+
+
+def mean_absolute_deviation(data, ead, predicted_lgd, realised_lgd):
+    """
+    Mean Absolute Deviation is a measure of the difference between the expected loss and
+    the realised loss. It is a measure of the loss that would have been incurred if the
+    actual losses were equal to the expected losses.
+    """
+
+    return np.sum(np.abs(data[realised_lgd] - data[predicted_lgd]) * data[ead]) / data[ead].sum()
+
+
+def elbe_t_test(df, lgd, elbe):
+    """
+    # df should contain the facilities for which backtesting will be performed.
+    # LGD is the observed LGD, ELBE is the ELBE for each facility, dataframe
+    # columns should be defined in this manner
+    """
+
+    N = len(df)
+    error = df[lgd] - df[elbe]
+    mean_error = error.mean()
+    num = np.sqrt(N) * mean_error
+    s2 = (((df[lgd] - df[elbe]) - mean_error) ** 2).sum() / (N - 1)
+    t_stat = num / np.sqrt(s2)
+    p_value = 2 * (1 - t.cdf(abs(t_stat), df=N - 1))
+
+    return pd.DataFrame({"facilities": [N], "lgd_mean": [df[elbe].mean()], "t_stat": [t_stat], "p_value": [p_value]})
+
+
+def normal_test(predicted_pd, realised_pd, alpha=0.05):
+    """
+    - Numeric vector of calibrated probabilities of default (PD).
+    - Numeric vector of observed default rates.
+    - alpha Significance level of p-value for implemented tests
+
+    The Normality test is a test of the null hypothesis that the residuals of a
+    regression are normally distributed.
+
+    The normal test is an approach to deal with the dependence problem that occurs in
+    the case of the binomial and chi-square tests. The normal test is a multi-period
+    test of correctness of a default probability forecast for a single rating category.
+
+    It is applied under the assumption that the mean default rate does not vary too
+    much over time and that default events in different years are independent. The
+    normal test is motivated by the Central Limit Theorem and is based on a normal
+    approximation of the distribution of the time-averaged default rates.
+
+    Simulation studies show that the quality of the normal approximation is moderate
+    but exhibits a conservative bias. As a consequence, the true Type I error tends
+    to be lower than the nominal level of the test, i.e. the proportion of
+    erroneous rejections of PD forecasts will be smaller than might be expected from
+    the formal confidence level of the test. The test seems even to be, to a certain
+    degree, robust against a violation of the assumption that defaults are independent
+    over time. However, the power of the test is moderate, in particular for short
+    time series (for example five years).
+
+    todo
+
+    """
+
+    length = len(predicted_pd)
+
+    standard_error = sum((predicted_pd - realised_pd) ** 2) - ((sum(predicted_pd - realised_pd) ** 2 / length)) / (
+        length - 1
+    )
+    t_stat = sum(predicted_pd - realised_pd) / np.sqrt(standard_error * length)
+    p_value = t.cdf(abs(t_stat), df=length)
+    estimate = sum(predicted_pd - realised_pd)
+
+    # create dataframe from inputs
+    return pd.DataFrame({"estimate": [estimate], "t_stat": [t_stat], "p_value": [p_value], "outcome": [np.nan]})
